/* Copyright 1992 John Bovey, University of Kent at Canterbury.
 *
 *  You can do what you like with this source code as long as
 *  you don't try to make money out of it and you include an
 *  unaltered copy of this message (including the copyright).
 */
/*
 * This module has been very heavily modified by R. Nation
 * (nation@rocket.sanders.lockheed.com).
 * No additional restrictions are applied
 *
 * Additional modification by Garrett D'Amore (garrett@netcom.com) to
 * allow vt100 printing.  No additional restrictions are applied.
 * Likewise mods by Charles Read for tk use- no additional restrictions
 * As usual, the author accepts no responsibility for anything, nor does
 * he guarantee anything whatsoever.
 */

#include <stdarg.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <grp.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <ctype.h>
#include <tk.h>
#ifdef ALPHA
#define FREEBSD
#endif
#ifdef AIXV3
#include <sys/select.h>
#endif   
#ifdef SVR4
#include <sys/stropts.h>
#define _NEW_TTY_CTRL
#endif
#ifdef AIXV3
#include <termio.h>
#else
#include <termios.h>
#endif /* AIXV3 */

#ifdef FREEBSD
#include <sys/ioctl.h>
#endif

#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>

#include "rxvt.h"
#include "command.h"
/* 
      #include "screen.h"
      #include "xsetup.h"
      #include "sbar.h"
*/
#if defined(VDISCRD) && !defined(VDISCARD)
#define	VDISCARD	VDISCRD
#endif

#if defined(VWERSE) && !defined(VWERASE)
#define	VWERASE		VWERSE
#endif

extern char *window_name;
extern char *icon_name;

/* for greek keyboard mapping */
#ifdef GREEK_KBD	
#include "grkelot.h"

static greek_mode_ON = False;
#endif

int size_set = 0;      /* flag set once the window size has been set */

#define KBUFSIZE	256	/* size of keyboard mapping buffer */
#define COM_BUF_SIZE   2048	/* size of buffer used to read from command */
#define STRING_MAX      512
#define ARGS_MAX         20     /* maximum # of args for esc sequences */

/*  Special character returned by get_com_char().
 */
#define GCC_NULL	0x100		/* Input buffer is empty */
#define ESC		033

/*  Flags used to control get_com_char();
 */
#define BUF_ONLY	1

/*  Global variables that are set up at the beginning and then not changed
 */
extern Display		*display;
extern Window		vt_win;
extern Window		main_win;
extern struct sbar_info sbar;

extern int login_shell;
extern int MetaHandling;
#ifdef MAPALERT
extern int map_alert;
KeySym AlertKeysym = XK_i;
#endif

int comm_fd = -1;/* file descriptor connected to the command */
static int comm_pid;	/* process id if child */
static int x_fd;	/* file descriptor of the X server connection */
static int fd_width;	/* width of file descriptors being used */
static int app_cur_keys = 0;/* flag to set cursor keys in application mode */
static int app_kp_keys = 0; /* flag to set application keypad keys */
#ifndef SVR4
static char ptynam[] = "/dev/ptyxx";
char ttynam[] = "/dev/ttyxx";
#else
char *ttynam, *ptsname();
#endif
static Atom wm_del_win;


extern int fat_sbar;
/* extern int console; */

KeySym SecureKeysym = XK_s;
KeySym BigFontKeysym = XK_greater;
KeySym SmallFontKeysym = XK_less;
KeySym PageUpKeysym = XK_Prior;
KeySym PageDownKeysym = XK_Next;
#ifdef GREEK_KBD
KeySym GreekSwitchKeysym = XK_Mode_switch;
int GreekMode = GREEK_ELOT928;
#endif
#ifdef PRINT_PIPE
KeySym PrintScreenKeysym = XK_Print;
#endif

int refresh_nl_count=0;
int refresh_nl_limit = 1;
extern int refresh_type;

/*  Terminal mode structures.
 */
static struct termios ttmode;
#ifndef FREEBSD
#ifndef CINTR
#ifndef _POSIX_VDISABLE
#define _POSIX_VDISABLE 0
#endif

#undef CTRL
#define CTRL(c) ((c) - '@')

#ifdef SVR4
#define CINTR 0177
#define CQUIT CTRL('U')
#define CERASE CTRL('H')
#else
#define CINTR CTRL('C')
#define CQUIT CTRL('\\')
#define CERASE 0177
#endif

#define CEOF CTRL('D')
#define CKILL CTRL('U')
#define CEOL _POSIX_VDISABLE
#define CEOL2 _POSIX_VDISABLE
#define CNSWTCH _POSIX_VDISABLE
#define CSTART CTRL('Q')
#define CSTOP CTRL('S')
#define CSUSP CTRL('Z');
#define CDSUSP _POSIX_VDISABLE
#define CRPRNT CTRL('R')
#define CFLUSH CTRL('O')
#define CWERASE CTRL('W')
#define CLNEXT CTRL('V')
#endif
#endif

#ifndef CEOL
#define CEOL _POSIX_VDISABLE
#endif

/*  Variables used for buffered command input.
 */
static unsigned char com_buf[COM_BUF_SIZE];
static unsigned char *com_buf_next, *com_buf_top;
unsigned char mask = 0xff;

static void catch_child(int);
static void catch_sig(int);
static int run_command(unsigned char *,unsigned char **);
static unsigned char *lookup_key(XEvent *,int *);
static int get_com_char(int);
void get_X_event(void);
void process_string(int);
#ifdef PRINT_PIPE
void process_print_pipe(void);
void process_print_screen(int);
#endif
void process_escape_sequence(void);
void process_csi_sequence(void);
void process_xterm_sequence(void);
void process_terminal_mode(int,int,int,int *);
void process_sgr_mode(int,int,int,int *);
struct stat ttyfd_stat;
int ttyfd;

#ifdef SVR4
#define _NEW_TTY_CTRL
#include <sys/resource.h>
int static getdtablesize() 
{
  struct rlimit   rlp;
  getrlimit(RLIMIT_NOFILE, &rlp);
  return(rlp.rlim_cur);
}
#endif


/*  Catch a SIGCHLD signal and exit if the direct child has died.
 */
static void catch_child(int nonsense)
{
  if (wait((int *)NULL) == comm_pid)
    {
/*      clean_exit(0);   */
/*        exit(0);        */
    }
      
}

char * SIGPIPED="0";
char *SIGREP()
{ char *s;
  s=SIGPIPED;SIGPIPED="0";
  return s; }
/*  Catch a fatal signal and tidy up before quitting
 */
static void catch_sig(int sig)
{ if (sig != SIGPIPE) {
  signal(sig,SIG_DFL);
/*  cleanutent();   */
  setuid(getuid());
    kill(getpid(),sig);};
  SIGPIPED="1"; 
}

/*  Run the command in a subprocess and return a file descriptor for the
 *  master end of the pseudo-teletype pair with the command talking to
 *  the slave.
 */
static int run_command(unsigned char *command,unsigned char **argv)
{
  int ptyfd;
  int uid, gid,shpgrp;
  unsigned char *s3, *s4;
  int i;
  int width, height;
#ifndef SVR4
  static char ptyc3[] = "pqrstuvwxyz";
  static char ptyc4[] = "0123456789abcdef";
 #endif
  unsigned char argv0[256];

  /*  First find a master pty that we can open.  
   */

#ifdef SVR4
  ptyfd = open("/dev/ptmx",O_RDWR);
  if (ptyfd < 0) 
    {
      fputs("Can't open a pseudo teletype",stderr);
      return(-1);
    }
  grantpt(ptyfd);
  unlockpt(ptyfd);
  fcntl(ptyfd,F_SETFL,O_NDELAY);
  ttynam=ptsname(ptyfd);
#else
  ptyfd = -1;
  for (s3 = ptyc3; *s3 != 0; s3++) 
    {
      for (s4 = ptyc4; *s4 != 0; s4++) 
	{
	  ptynam[8] = ttynam[8] = *s3;
	  ptynam[9] = ttynam[9] = *s4;
	  if ((ptyfd = open(ptynam,O_RDWR)) >= 0) 
	    {
	      if (geteuid() == 0 || access(ttynam,R_OK|W_OK) == 0)
		break;
	      else 
		{
		  close(ptyfd);
		  ptyfd = -1;
		}
	    }
	}
      if (ptyfd >= 0)
	break;
    }
  if (ptyfd < 0) 
    {
      fputs("Can't open a pseudo teletype",stderr);
      return(-1);
    }
  fcntl(ptyfd,F_SETFL,O_NDELAY);
#endif  
  for (i = 1; i <= 15; i++)
    signal(i,catch_sig);
  signal(SIGCHLD,catch_child);
  lstat(ttynam,&ttyfd_stat);

/*  makeutent(&ttynam[5]);	 stamp /etc/utmp    */

  comm_pid = fork();
  if (comm_pid < 0) 
    {
      fputs("Can't fork",stderr);
      return(-1);
    }
  if (comm_pid == 0) 
    {
      struct group *gr;
      
/* FIXME: I (ah) don't know whether unsetenv is defined by these */
/* I don't think its very portable: I'll use #ifdef UNSETENV
 * RN */
#ifdef UNSETENV
      /*
       * bash may set these env vars, but they should be clear for the child
       * otherwise the old settings are passed on and will confuse term size 
       */
       unsetenv("LINES");
       unsetenv("COLUMNS");
#endif 
      if ((ttyfd = open(ttynam,O_RDWR)) < 0) 
	{
	  fputs("could not open slave tty ",stderr);
	 /* clean_exit(1); */
            exit(1);
	}
#ifdef SVR4
      ioctl(ttyfd,I_PUSH,"ptem");
      ioctl(ttyfd,I_PUSH,"ldterm");
#endif

      uid = getuid();
#ifndef SVR4
      if ((gr = getgrnam("tty")) != NULL)
	gid = gr->gr_gid;
      else
	gid = -1;
      fchown(ttyfd,uid,gid);
      fchmod(ttyfd,0600);
#endif
#ifdef TIOCCONS
/*     if (console) 
	{
	  int on = 1;
	  if (ioctl (ttyfd, TIOCCONS, (unsigned char *)&on) == -1)
	    fprintf(stderr, "xvt: cannot open console\n");
	} */
#endif  /* TIOCCONS */
#ifdef HPUX
      for (i = 0; i < sysconf(_SC_OPEN_MAX); i++)
	if (i != ttyfd)
	  close(i);
#else
      for (i = 0; i < getdtablesize(); i++)
	if (i != ttyfd)
	  close(i);
#endif

      dup(ttyfd);
      dup(ttyfd);
      dup(ttyfd);

      if (ttyfd > 2)
	close(ttyfd);
      
      if (setsid() < 0)
	perror("failed to set process group");
#if defined(TIOCSCTTY)
      ioctl(0, TIOCSCTTY, 0) ;
#endif
      {
	int pgrp = getpid();
	ioctl(0, TIOCSPGRP, (char *)&pgrp);
	setpgrp();
	close(open(ttynam, O_WRONLY, 0));
	setpgrp ();
      }

      /* init of termios structure		*/
#ifdef FREEBSD
      ioctl(0,TIOCGETA,(char *)&ttmode);
#else
#   ifdef HPUX
      tcgetattr(0, &ttmode);
#   else
      ioctl(0,TCGETS,(char *)&ttmode);
#   endif        
#endif
 
#ifdef HPUX
      ttmode.c_iflag = BRKINT | IGNPAR | ICRNL| IXON;
      ttmode.c_lflag = ISIG|IEXTEN|ICANON|ECHO|ECHOE|ECHOK;
#else
      ttmode.c_iflag = BRKINT | IGNPAR | ICRNL| IXON | IMAXBEL;
      ttmode.c_lflag = ISIG|IEXTEN|ICANON|ECHO|ECHOE|ECHOK|ECHOCTL|ECHOKE;
#endif                                                       
      ttmode.c_oflag = OPOST | ONLCR ;
      ttmode.c_cflag = B9600 | CS8 | CREAD;

      ttmode.c_cc[VEOF] = CEOF;
#ifdef ALPHA
      (unsigned) ttmode.c_cc[VEOL] = CEOL;
#else
      ttmode.c_cc[VEOL] = CEOL; 
#endif
      ttmode.c_cc[VINTR] = CINTR;
      ttmode.c_cc[VQUIT] = CQUIT;
      ttmode.c_cc[VERASE] = CERASE;
      ttmode.c_cc[VKILL] = CKILL;
#ifdef HPUX
      ttmode.c_cc[VSUSP] = CSWTCH;
#else
      ttmode.c_cc[VSUSP] = CSUSP;
#endif
#ifdef VDSUSP
      ttmode.c_cc[VDSUSP] = CDSUSP;
#endif
      ttmode.c_cc[VSTART] = CSTART;
      ttmode.c_cc[VSTOP] = CSTOP;
#ifdef VREPRINT
      ttmode.c_cc[VREPRINT] = CRPRNT;
#endif
#ifdef VDISCARD
      ttmode.c_cc[VDISCARD] = CFLUSH;
#endif
#ifdef VWERASE
      ttmode.c_cc[VWERASE] = CWERASE;
#endif
#ifdef VLNEXT
      ttmode.c_cc[VLNEXT] = CLNEXT;
#endif
#ifdef VSWTC
      ttmode.c_cc[VSWTC] = 0;
#endif                           

#ifdef VSWTCH
      ttmode.c_cc[VSWTCH] = 0;
#endif 
#if VMIN != VEOF
      ttmode.c_cc[VMIN] = 1;
#endif
#if VTIME != VEOL
      ttmode.c_cc[VTIME] = 0;
#endif

      if(mask == 0x7f)
	ttmode.c_cflag = B9600 | PARENB | CS7 | CREAD;
#ifdef FREEBSD
      ioctl(0,TIOCSETA,(char *)&ttmode);
#else
#   ifdef HPUX
      tcsetattr(0, TCSANOW, &ttmode);
#   else
      ioctl(0,TCSETS,(char *)&ttmode);
#   endif        
#endif
/*      scr_get_size(&width,&height); */
/*      tty_set_size(0,width,height);  */
        tty_set_size(0,80,24);
/*      if(login_shell)
	{
          char *tmp;
	  strcpy(argv0,"-");
          if ((tmp = strrchr(argv[0], '/')) == NULL)
            tmp = argv[0];
          else
            tmp++;
	  strcat(argv0,tmp);
	  argv[0] = argv0;
	}          
*/
      setgid(getgid());
      setuid(uid);

      execvp(command,(char **)argv);
      fputs("Couldn't execute",stderr);
    /*  clean_exit(1); */
           exit(1);
    }
  return(ptyfd);
}

/*  Tell the teletype handler what size the window is.  Called after a window
 *  size change.
 */
void tty_set_size(int fd,int width,int height)
{
  struct winsize wsize;
  
  if(fd < 0)
    return;

  wsize.ws_row = (unsigned short)height;
  wsize.ws_col = (unsigned short)width;
  ioctl(fd,TIOCSWINSZ,(char *)&wsize);
}


#ifdef NO_SETSID
setsid()
{
  if (setpgrp(0,0) < 0) return(-1);
  return(0);
}
#endif

char ans[200];
int runcommand(ClientData cd, Tcl_Interp *interp, int argc, char *argv[])
{ int fd;
  if (argc<2)  {interp->result="Usage: runcommand command [args...]";
       return TCL_ERROR;};  
  fd=run_command(argv[1],&argv[1]);
  if (fd>=0){
  sprintf(ans,"%d file%d",comm_pid,fd);
  interp->result=ans;
  return TCL_OK;};
  interp->result="could not run command for some reason";
  return TCL_ERROR;
}







